from git import RemoteProgress
from tqdm import tqdm


class ClonedProgreeBar(RemoteProgress):
    def __init__(self, i, name) -> None:
        super().__init__()
        self.pbar = tqdm(
            position=i,
            desc=f"Downloading:{name}",
            bar_format="{l_bar}{bar:50}| {n_fmt}/{total_fmt} [{elapsed}]",
        )

    def update(
        self,
        op_code: int,
        cur_count: str | float,
        max_count: str | float | None = None,
        message: str = "",
    ) -> None:
        self.pbar.total = max_count
        self.pbar.n = cur_count
        self.pbar.refresh()



import xml.etree.ElementTree as ET
import re
import json


def extract_git_links_xml(path):
    root = ET.parse(path)

    links = []
    git_pattern = re.compile(r"https?://(?:www\.)?github\.com/[^\s]+")

    for elem in root.iter():
        if elem.text:
            match = git_pattern.search(elem.text)
            if match:
                links.append(match.group(0))

    return links

def extract_links_lineby_line(path):
    with open(path,"r") as f:
        file=f.readlines()
    
    return list(file)


def extract_git_links_json(path):
    def search(doc, links):
        if isinstance(doc, dict):
            for key, data in doc.items():
                search(data, links)
        elif isinstance(doc, list):
            for data in doc:
                search(data, links)
        elif isinstance(doc, str):
            match = git_pattern.search(doc)
            if match:
                links.append(match.group(0))

    ext = path.split(".")[-1]
    with open(path, "r") as file:
        if ext.lower() == "json":
            doc = json.load(file)
        else:
            doc = [json.loads(line) for line in file]

    links = []

    git_pattern = re.compile(r"https?://(?:www\.)?github\.com/[^\s]+")
    search(doc, links)
    return links



class SSH_key_not_set(Exception):
    def __init__(self, *args: object) -> None:
        self.message="SSH Key file not set"
        super().__init__(self.message)

class Permission_Denied_SSH(Exception):
    def __init__(self, *args: object) -> None:
        self.message="Permission Denied Check if the set SSH  jey is correct"
        super().__init__(self.message)



import configparser
from typing import List
from git import Repo
import os
import requests
import subprocess
from concurrent.futures import ThreadPoolExecutor


from .ProgressBar import ClonedProgreeBar
from .Zipper import Zipper
from .Generator import Generator

from .Parsers import extract_git_links_xml, extract_git_links_json,extract_links_lineby_line

from .errors import SSH_key_not_set,Permission_Denied_SSH
import atexit
config=configparser.ConfigParser()
config_path = os.path.join(os.path.dirname(__file__), 'conf.cfg')
config.read(config_path)

class GitToMark:
    BASE_DIR = "./RepoStore"
    OUT_DIR = "./Output"
    
    def _cleanup(self):
        """Clean up Git resources properly"""
        for repo in self.repo:
            if repo and hasattr(repo.git, '_terminate'):
                repo.git._terminate()
        self.repo.clear()
    def __init__(self, urls: List["str"]|str,ssh=False) -> None:
        '''
        urls:List of git repo links
        ssh:Enable fetching from SSH requires setting the ssh key path in GitToMark Config 
        '''
        
        self.ssh=ssh
        if ssh:
            if not config["Default"]["Shh_key_path"]:
                raise SSH_key_not_set
            x=subprocess.run(f'ssh -i "{config["Default"]["Shh_key_path"]}" -T git@github.com',text=False)
            if x.returncode!=1:
                raise Permission_Denied_SSH
           



        if isinstance(urls,str):
            urls=[urls]
        self.urls = []
        self.repo = []
        self.UID = [self.extract_repo_name(url,ssh) for url in urls]
        atexit.register(self._cleanup)
        with ThreadPoolExecutor(int(config["Default"]["Threads"])) as DownloadPool:
            # self.repo=[self.Create_repo_Handle(url,i) for i,url in enumerate(self.urls)]
            for r in DownloadPool.map(
                lambda args: self.Create_repo_Handle(*args,ssh), enumerate(urls)
            ):
                if r:
                    self.repo.append(r)
    
    @classmethod
    def from_xml(cls, path):
        links = extract_git_links_xml(path)
        return cls(links)

    @classmethod 
    def from_file(cls,path):
        links=extract_links_lineby_line(path)
        return cls(links)
    @classmethod
    def from_json(cls, path):
        links = extract_git_links_json(path)
        return cls(links)
    @staticmethod
    def config(threads=10,ssh_path=""):
        config["Default"]["THREADS"]=str(threads)
        config["Default"]["Shh_key_path"]=ssh_path
    @property
    def store(self):
        Zipper(self.UID)

    def extract_repo_name(self, path,ssh):
        if not ssh:
            name = path.split("/")[-2:]
            name[1] = name[1][:-4]
            name.insert(1, "-")
            name = "".join(name)
            return name
        else:
            name=path.split(":")[1].split(".")[0].split("/")
            name.insert(1,"-")
            name="".join(name)
            return name

    @property
    def generate(self):
        with ThreadPoolExecutor(int(config["Default"]["THREADS"])) as pool:
            i = 0
            for url, repo in zip(self.urls, self.repo):
                if self.ssh:
                    url=url.split(":")[1:]
                    url=''.join(url)

                Reponame = url.split("/")[-2:]
                Reponame[1] = Reponame[1][:-4]
                outfile = Reponame[1]
                Reponame.insert(1, "-")
                Reponame = "".join(Reponame)
                # pool.submit(Generator,(Reponame,repo,outfile))
                instance = Generator(Reponame, repo, outfile, i)

                    
                pool.submit(instance.Generate_MD)

                i += 1

    def Create_repo_Handle(self, i, url,ssh):
        handle = None
        try:
            if os.path.exists(self.BASE_DIR + f"/{self.UID[i]}"):
                self.urls.append(url)
                handle = Repo(self.BASE_DIR + f"/{self.UID[i]}")
            else:
                if not ssh:
                    if requests.get(url).status_code == 200:
                        handle = Repo.clone_from(
                            url,
                            self.BASE_DIR + f"/{self.UID[i]}",
                            progress=ClonedProgreeBar(i, self.UID[i]),
                        )
                        self.urls.append(url)
                    else:
                        print(f"Invalid Url:{url}")
                else:
                    handle = Repo.clone_from(
                        url,
                        self.BASE_DIR + f"/{self.UID[i]}",
                        progress=ClonedProgreeBar(i, self.UID[i]),
                        env={"GIT_SSH_COMMAND": f'ssh -i "{config["Default"]["Shh_key_path"]}"'}
                    )
                    self.urls.append(url)
            
            if handle:
                self.repo.append(handle)
            return handle
            
        except Exception as e:
            print(f"Error creating repo handle: {e}")
            return None


import json
import os
from pathlib import Path

class Markdown:
    EXT_MAP = json.load(open(Path(__file__).parent/"extensions.json", "r"))["ext"]

    def __init__(self, filename) -> None:
        if not os.path.exists("./Output"):
            os.makedirs("./Output")
        self.handle = open(filename, "w", encoding="utf-8")

    def add_header(self, content, level=1):
        if level > 6:
            raise ValueError("Header level must be between 1 and 6.")
        self.handle.write("#" * level + " " + content + "\n")

    def add_code_block(self, content, fileext):
        ext = ""
        if fileext.lower() in self.EXT_MAP:
            ext = self.EXT_MAP[fileext.lower()]
        ctx = f"""\n```{ext}\n{content}\n```\n"""
        self.handle.write(ctx)

    def add_para(self, content):
        self.handle.write(content)



from sys import stdout
import os, platform
from git import Repo
from tqdm import tqdm
from .MarkdownUtils import Markdown
import pathlib
import os


class Generator:
    BASE_DIR = "./RepoStore"
    OUT_DIR = pathlib.Path("Output")

    def __init__(self, reponame: str, repo: Repo, outfile: str, i):
        self.out_dir = pathlib.Path(os.getcwd()) / self.OUT_DIR
        self.repoName = reponame
        self.repo = repo
        self.tree = repo.head.commit.tree
        self.MDout = str(self.out_dir / f"{reponame}.md")
        self.MD = Markdown(self.MDout)
        self.total = 0
        self.dirs = 0
        if  os.stat(self.OUT_DIR / f"{self.repoName}.md").st_size == 0:
            self.pbar = tqdm(
                position=i,
                desc=f"Processing:{reponame}",
                bar_format="{l_bar}{bar:50}| {n_fmt}/{total_fmt} [{elapsed}]",
            )
        # self.Generate_MD()

    def print_files_from_git(self, root, level=0):
        for entry in root:
            # print(f'{"-" * 4 * level}| {entry.path}, {entry.type}')
            self.total += 1

            tmp = f"{' ' * level * 2}- {entry.path}" + "\n"

            yield tmp
            if entry.type == "tree":
                self.dirs += 1
                yield from self.print_files_from_git(entry, level + 1)

    def write_files_in_MD(self, root, level=0):
        for entry in root:
            if entry.type == "tree":
                self.pbar.update(1)
                self.write_files_in_MD(entry, level + 1)
            else:
                self.pbar.update(1)
                self.MD.add_header(entry.path, 2)
                with open(
                    self.BASE_DIR + "/" + self.repoName + "/" + entry.path,
                    "r",
                    encoding="utf-8",
                ) as f:
                    x = entry.path.split(".")[-1]
                    if x.lower() != "md":
                        self.MD.add_code_block(f.read(), x)

    def __clear_stdout(self):
        if platform.system() == "Windows":
            os.system("cls")
        else:
            os.system("clear")

    def write_git_tree(self):
        self.MD.add_header("Git Tree", 2)

        for line in self.print_files_from_git(self.tree):
            # stdout.write(f"\033[{self.i+1};1H")
            # stdout.write("\033[2K")
            # stdout.write(f"\rFiles:{self.total-self.dirs}|Dirs:{self.dirs}")
            # stdout.flush()
            # self.__clear_stdout()
            # pbar.total=self.total-self.dirs

            self.MD.add_para(
                line,
            )
        self.pbar.total = self.total

    def write_header(self):
        # split=self.urls.split('/')[-2:]
        # split.insert(1,"/")
        #
        # Headinfo=''.join(split)[:-4]
        self.MD.add_header(self.repoName.replace("-", "/"))

    def Generate_MD(self):
        if (
            os.path.exists(self.OUT_DIR / f"{self.repoName}.md")
            and os.stat(self.OUT_DIR / f"{self.repoName}.md").st_size != 0
        ):
            print(f"{self.repoName} Already Exists")
            return
        self.write_header()
        self.write_git_tree()

        self.write_files_in_MD(self.tree)

    def print_tree(self):
        self.print_files_from_git(self.tree)



from GitToMarkdown import GTM
from GitToMarkdown import Generator
from GitToMarkdown import MarkdownUtils
from GitToMarkdown import Parsers
from GitToMarkdown import ProgressBar
from GitToMarkdown import Zipper
from GitToMarkdown import errors


from GitToMarkdown.GTM import (GitToMark, config, config_path,)
from GitToMarkdown.Generator import (Generator,)
from GitToMarkdown.MarkdownUtils import (Markdown,)
from GitToMarkdown.Parsers import (extract_git_links_json,
                                   extract_git_links_xml,
                                   extract_links_lineby_line,)
from GitToMarkdown.ProgressBar import (ClonedProgreeBar,)
from GitToMarkdown.Zipper import (Zipper,)
from GitToMarkdown.errors import (Permission_Denied_SSH, SSH_key_not_set,)

__all__ = ['ClonedProgreeBar', 'GTM', 'Generator', 'GitToMark', 'Markdown',
           'MarkdownUtils', 'Parsers', 'Permission_Denied_SSH', 'ProgressBar',
           'SSH_key_not_set', 'Zipper', 'config', 'config_path', 'errors',
           'extract_git_links_json', 'extract_git_links_xml',
           'extract_links_lineby_line', 'tempCodeRunnerFile']


import zipfile
from zipfile import ZipFile
import hashlib
#hello

class Zipper:
    OUT_DIR = "./Output"

    def __init__(self, reponame) -> None:
        self.reponames = reponame
        self.UID = hashlib.sha256(("".join(self.reponames)).encode("utf-8")).hexdigest()
        self.archive()

    def archive(self):
        with ZipFile(f"Zips/{self.UID}.zip", "w") as zfile:
            for entry in self.reponames:
                zfile.write(self.OUT_DIR + f"/{entry}.md", entry + ".md")
